import * as React from 'react';
import {observer} from "mobx-react-lite";
import {DashboardComponentModel} from "../../../stores/DashboardComponentStore";
import {LineChart} from "@rx-suite/ui";
import {useGraphOptions} from "../../../hooks/useGraphOptions";
import {t} from "@lingui/macro";
import {
    MACHINE_PERFORMANCE_OVERVIEW_COLORS,
    resolveMachinePerformanceSeriesName
} from "../machine_performance_utils/machinePerformanceOverviewUtilities";
import {useMachinePerformanceOverviewLegend} from "../machine_performance_utils/useMachinePerformanceOverviewLegend";

export type DashboardComponentPerformanceChartProps = {
    component: DashboardComponentModel
};

type ThroughputMetric = 'jobs' | 'lens';

export const DashboardComponentPerformanceChart: React.FunctionComponent<DashboardComponentPerformanceChartProps> = observer(({component}) => {
    const data = component.viewModel.data;
    const options = useGraphOptions(component);
    const UPTIME_SERIES_INDEX = 1;
    const UTILIZATION_SERIES_INDEX = 0;

    const uptimeValueForDataPoint = React.useCallback((dataPointIndex: number) => {
        // sum the uptime and the processing of the series to get the utilization
        return data.series[0]['data'][dataPointIndex].y + data.series[1]['data'][dataPointIndex].y;
    },[data.series]);

    const metric = React.useMemo(() => {
        if (!data || !data.series || !data.series[2] || !data.series[2].metric) {
            return t`Jobs`;
        }
        const metric: ThroughputMetric = data.series[2].metric;
        return {
            jobs: t`Pairs`,
            lens: t`Lens`
        }[metric];
    }, [data]);

    const seriesPointValue = React.useCallback((value, {seriesIndex, dataPointIndex}) => {
        if(seriesIndex === UPTIME_SERIES_INDEX){
            value = uptimeValueForDataPoint(dataPointIndex);
        }

        if(seriesIndex == UTILIZATION_SERIES_INDEX || seriesIndex == UPTIME_SERIES_INDEX){
            return `${value.toFixed(2)}%`;
        }

        return `${value} ${metric}`;
    }, [metric, uptimeValueForDataPoint]);

    const throughputSum = React.useMemo(() => {
        if (!data || !data.series || !data.series[2]) {
            return 0;
        }
        return data.series[2].data.map(d => Number(d.y)).reduce((a: number, b: number) => a + b, 0);
    }, [data]);
    
    const updatePerformanceOptions = React.useCallback((options) => {
        options.subtitle = {
            text: `Total ${metric}: ${throughputSum}`,
            align: 'left',
            offsetX: 64,
            style: {
                fontSize: '14px',
                fontWeight: 'bold',
                color: 'black'
            }
        };

        options.chart = {
            stacked: true,
            toolbar: {
                show: false
            },
            zoom: {
                enabled: false
            }
        };

        options.tooltip = {
            y: {
                formatter: seriesPointValue,
                title: {
                    formatter: resolveMachinePerformanceSeriesName
                },
            },
            fillSeriesColor: false,
            marker: false
        };

        options.legend = {
            position: 'bottom',
            markers: {
                width: 0
            },
            formatter: useMachinePerformanceOverviewLegend
        };

        options.xaxis = {
            type: 'category'
        };

        options.dataLabels = {
            enabled: true,
            enabledOnSeries: [2]
        };

        options.stroke = {
            width: 2
        };

        options.yaxis = [
            {
                min: 0,
                max: 100,
                floating: false,
                tickAmount: 5,
                forceNiceScale: false,
                decimalsInFloat: 0,
                seriesName: 'waitingforjobs',
                title: {
                    text: t`Utilization/Uptime (%)`
                },
            },
            {
                show: false,
                min: 0,
                max: 100,
                floating: false,
                tickAmount: 5,
                seriesName: 'processing',
                forceNiceScale: false,
            },
            {
                seriesName: 'throughput',
                opposite: true,
                tickAmount: 5,
                decimalsInFloat: 0,
                title: {
                    text: `Throughput` + `(${metric})`
                }
            }
        ];

        options.colors = [
            MACHINE_PERFORMANCE_OVERVIEW_COLORS.utilization.over,
            MACHINE_PERFORMANCE_OVERVIEW_COLORS.uptime.over,
            MACHINE_PERFORMANCE_OVERVIEW_COLORS.throughput
        ];

    }, [metric, seriesPointValue, throughputSum]);

    if (component.isNew) {
        const fakeKeys = [...Array(3).keys()];
        const fakeSeries = [
            {
                type: 'bar',
                data: fakeKeys.map((v, i) => ({
                    x: i + 1,
                    y: Math.floor(Math.random() * 50)
                }))
            },
            {
                type: 'bar',
                data: fakeKeys.map((v, i) => ({
                    x: i + 1,
                    y: Math.floor(Math.random() * 50)
                }))
            },
            {
                type: 'line',
                data: fakeKeys.map((v, i) => ({
                    x: i + 1,
                    y: Math.floor(Math.random() * 30)
                }))
            }
        ];

        options.chart = { stacked: true };
        options.legend = { position: "bottom" };
        options.xaxis = { title: { text: undefined }};
        options.yaxis = {
            min: 0,
            max: 100,
            floating: false,
            tickAmount: 5,
            forceNiceScale: false,
            decimalsInFloat: 0,
            title: { text: undefined }
        };

        return <LineChart series={fakeSeries}
                          options={{
                              ...options,
                              colors: ['#ABABAB', '#DCDCDC', '#565656']
                          }}
        />;
    } else {
        updatePerformanceOptions(options); // inline updates

        return <LineChart series={data.series || []}
                          options={options}
        />;
    }
});
