import {attribute, DataStore, isArrayLike, Model, model, Transport, Types} from '@rx-suite/data';
import {t} from "@lingui/macro";
import * as yup from 'yup';
import {action, computed, observable, override, runInAction, when} from "mobx";
import {DashboardTargetModel, DashboardTargetStore} from "./DashboardTargetStore";
import {computedFn} from "mobx-utils";
import {Annotation, AvailableCategory, AvailableSeries, CustomTarget, ReturnedCategory, ReturnedSeries} from "./Types";
import {v4 as uuidv4} from 'uuid';
import {ApexOptions} from "apexcharts";
import {
    getDataLabelFormatter,
    getLegendFormatter,
    getTooltipFormatter,
    getXAxisFormatter
} from "./formatters/DashboardComponentFormatters";

class DashboardComponentResultTransport<T extends Model> extends Transport<T> {
    async run(id: string) {
        const resp = await this.fetch(`/${id}/run`, {
            method: 'POST',
            headers: {
                'Content-Type': 'application/json'
            }
        });
        return resp.json();
    }
}

const hashString = function (str: string) {
    const len = str.length;
    const c1 = 0xcc9e2d51, c2 = 0x1b873593;
    let h1 = 8467; //arbitrary seed data
    const roundedEnd = len & ~0x1;
    const mul32 = (m: number, n: number) => {
        const nlo = n & 0xffff;
        const nhi = n - nlo;
        return ((nhi * m | 0) + (nlo * m | 0)) | 0;
    };

    for (let i = 0; i < roundedEnd; i += 2) {
        let k1 = str.charCodeAt(i) | (str.charCodeAt(i + 1) << 16);

        k1 = mul32(k1, c1);
        k1 = ((k1 & 0x1ffff) << 15) | (k1 >>> 17);  // ROTL32(k1,15);
        k1 = mul32(k1, c2);

        h1 ^= k1;
        h1 = ((h1 & 0x7ffff) << 13) | (h1 >>> 19);  // ROTL32(h1,13);
        h1 = (h1 * 5 + 0xe6546b64) | 0;
    }
    if ((len % 2) === 1) {
        let k1 = str.charCodeAt(roundedEnd);
        k1 = mul32(k1, c1);
        k1 = ((k1 & 0x1ffff) << 15) | (k1 >>> 17);  // ROTL32(k1,15);
        k1 = mul32(k1, c2);
        h1 ^= k1;
    }

    // finalization
    h1 ^= (len << 1);

    // fmix(h1);
    h1 ^= h1 >>> 16;
    h1 = mul32(h1, 0x85ebca6b);
    h1 ^= h1 >>> 13;
    h1 = mul32(h1, 0xc2b2ae35);
    h1 ^= h1 >>> 16;

    return h1;
};

const stringToColour = function (str: string) {
    const hash = hashString(str);
    let colour = '#';
    for (let i = 0; i < 3; i++) {
        const value = (hash >> (i * 8)) & 0xFF;
        colour += ('00' + value.toString(16)).substr(-2);
    }
    return colour;
}

@model({
    transportLayer: new DashboardComponentResultTransport("/api/dashboard_components", "dashboard_component"),
    permissionTarget: "DashboardComponent",
    routingInfo: {
        list: "/dashboard_components",
        create: "/dashboard_components/create",
        update: "/dashboard_components/:id/edit",
        read: "/dashboard_components/:id/read"
    }
})
export class DashboardComponentModel extends Model {
    types: Types = {
        parameters: 'json',
        display_options: 'json',
        data: 'json',
    };

    constructor() {
        super();
        super.setup();
    }

    @attribute({
        label: t`Name`,
        description: t`This is how you will identify the component in lists (for example, you can search for this name).  It will also be shown as a title on some component types.`,
        validation: yup.string().required(t`You must give your component a name`)
    })
    name = '';

    @attribute({
        label: t`Description`,
        description: t`Add any notes here.  The information will not be included in the dashboard display - it is for reference purposes only.`,
        validation: yup.string()
    })
    desc = '';

    @attribute({
        label: t`Run As`,
        description: t`Run the dashboard using this user's credentials`,
        validation: yup.string().nullable()
    })
    run_user_id = '';

    @attribute({
        label: t`Refresh Interval (Seconds)`,
        description: t`Control how often the component fetches fresh data.`,
        validation: yup.number().min(1, t`Choose a larger refresh interval`).required(t`You must give your component a refresh interval`)
    })
    refresh_interval = 60;

    @attribute({
        label: t`Parameters`,
        validation: yup.object().required(),
        autoTable: false
    })
    parameters: any = observable({});

    @attribute({
        label: t`Display Options`,
        validation: yup.object().nullable(),
        autoTable: false
    })
    display_options: any = observable({});

    @attribute({
        label: t`Display Type`,
        validation: yup.string().when('dashboard_target_id', (dashboard_target_id, schema) => {
            if (dashboard_target_id.toString()) {
                const target = (DashboardTargetStore.data.get(dashboard_target_id.toString()) || DashboardTargetStore.get(dashboard_target_id.toString())) as DashboardTargetModel;
                if (target && target.displayTypes.length) {
                    return schema.required(`You must choose a display type`);
                }
            }
            return schema;
        }),
        autoTable: false
    })
    display_type = '';

    @attribute({
        label: t`Data`,
        autoTable: false, autoForm: false
    })
    data: any = observable({});

    @attribute({
        label: t`Last Run Error`,
        validation: yup.string().nullable(),
        autoForm: false
    })
    last_run_error: string | null = null;

    @attribute({
        label: t`Target`,
        connects: 'DashboardTargetModel',
        description: t`The selected target determines what kinds of data can be fetched and what options you have to display it.  Available targets depend on your installed Rx-Suite modules.`,
        validation: yup.number().required(t`You must set a dashboard target`),
        autoTable: false
    })
    dashboard_target_id: number | undefined = undefined;

    @attribute({label: t`Created At`, autoForm: false})
    created_at: Date = new Date();

    @attribute({label: t`Updated At`, autoForm: false})
    updated_at: Date = new Date();

    @attribute({label: t`Queued`, autoForm: false})
    queued: boolean = false;

    @computed
    get isDisplayTypeOverride() {
        if (this.viewModel.dashboard_target) {
            return (this.viewModel.dashboard_target as DashboardTargetModel).isDisplayTypeOverride;
        }
        return false;
    }

    @computed
    get displayTypes() {
        if (this.viewModel.dashboard_target) {
            return (this.viewModel.dashboard_target as DashboardTargetModel).displayTypes;
        }
        return [];
    }

    @computed
    get displayTypeName() {
        const typeDef = this.displayTypes.find(t => t.value === this.display_type);
        if (typeDef) {
            return typeDef.label;
        } else {
            return this.display_type || ''
        }
    }

    isValidDisplayType = computedFn(function getDef(this: DashboardComponentModel, type: string): any {
        return this.displayTypes && (this.displayTypes.length > 0) && this.displayTypes.find(t => t.value === type);
    });

    isGroupIncludedInLabel(): boolean {
        return (this.viewModel.dashboard_target as DashboardTargetModel).ui_hints?.target_label_should_include_group
    }

    targetParameterName = computedFn(function getDef(this: DashboardComponentModel, value: string): any {
        if (this.viewModel.parameters && this.viewModel.parameters.custom_targets) {
            const custom = this.viewModel.parameters.custom_targets.find((c: CustomTarget) => c.id === value);
            if (custom) {
                return custom.label;
            }
        }

        const dashboardTarget = this.viewModel.dashboard_target as DashboardTargetModel;
        if (dashboardTarget && dashboardTarget.parameters && dashboardTarget.parameters.targets) {
            const returnDef = (dashboardTarget.parameters.targets.value_list || []).find(t => t.value === value);
            if (returnDef && returnDef.label) {
                return this.isGroupIncludedInLabel() ? `${returnDef.group} - ${returnDef.label}` : returnDef.label
            } else {
                return value;
            }
        } else {
            return value;
        }
    });

    targetParameterProp = computedFn(function getDef(this: DashboardComponentModel, value: string, prop: string): any {
        const dashboardTarget = this.viewModel.dashboard_target as DashboardTargetModel;
        if (dashboardTarget && dashboardTarget.parameters && dashboardTarget.parameters.targets) {
            const returnDef = (dashboardTarget.parameters.targets.value_list || []).find(t => t.value === value) as any;
            if (returnDef) {
                return (returnDef[prop] === undefined) ? null : returnDef[prop];
            } else {
                return null;
            }
        } else {
            return null;
        }
    });

    // "." confuses handleNestedAssignment, which tries to dig one step too many
    getGoalId(targetId: string) {
        return targetId.replace(".", "&|");
    }

    @action.bound
    resetParameters() {
        this.nestedAssignment('parameters', {});
        if (!this.isValidDisplayType(this.viewModel.display_type)) {
            if (this.displayTypes.length === 1) {
                this.nestedAssignment('display_type', this.displayTypes[0].value);
            } else {
                this.nestedAssignment('display_type', '');
            }
        }
        if (this.viewModel.dashboard_target) {
            Object.values((this.viewModel.dashboard_target as DashboardTargetModel).parameters).forEach(parameter => {
                if (parameter.default) {
                    runInAction(() => {
                        this.nestedAssignment(`parameters.${parameter.id}`, parameter.default)
                    });
                }
            });
        }
        this.nestedAssignment('data', {})
    }

    @action.bound
    pushAnnotation() {
        this.nestedAssignment('display_options.annotations', [...(this.viewModel.display_options?.annotations || []), {
            id: uuidv4(),
            label: "",
            type: "yaxis",
            value: 0,
            color: '#8B0000',
            thickness: 10,
            dashed: false
        } as Annotation])
    }

    @action.bound
    removeAnnotation(annotation: Annotation) {
        this.nestedAssignment('display_options.annotations', (this.viewModel.display_options?.annotations || []).filter((a: Annotation) => a !== annotation));
    }

    @action.bound
    pushCustomTarget() {
        this.nestedAssignment('parameters.custom_targets', [...((this.viewModel.parameters || {}).custom_targets || []), {
            id: uuidv4(),
            label: "",
            combines: [],
        } as CustomTarget])
    }

    @action.bound
    removeCustomTarget(target: CustomTarget) {
        this.nestedAssignment('parameters.custom_targets', ((this.viewModel.parameters || {}).custom_targets || []).filter((t: CustomTarget) => t !== target));
    }


    @computed get targetIds() {
        const params = this.viewModel.parameters || {};
        const targetIds = [...(params.targets || []),
            ...(params.custom_targets || []).map((c: CustomTarget) => c.id)
        ]
        if (targetIds.length == 0) {
            const targets = Object.values((this.viewModel.dashboard_target as DashboardTargetModel).parameters || []).find(x => x.id == "targets")?.value_list || []
            return targets.map((c: any) => c.value)
        }
        return targetIds
    }

    @override
    get exportName() {
        return this.name || this.getId() as string;
    }

    updateData = async () => {
        const transport = this.store.transportLayer as DashboardComponentResultTransport<DashboardComponentModel>;
        return transport.run(this.getId());
    }

    asExport() {
        const standardDump = super.asExport() as any;
        const target = this.dashboard_target as DashboardTargetModel;
        const result = {
            dashboard_target: {
                module_id: target.module_id,
                dashboard_id: target.dashboard_id,
            }
        } as any;
        Object.keys(standardDump).forEach(k => {
            if (k !== 'dashboard_target_id') {
                result[k] = standardDump[k]
            }
        });
        return result;
    }

    /**
     * Inverse of asExport
     * @returns {object} An export-friendly object
     */
    @override
    fromExport(json: Model) {
        this.updateFromJson(json, true);
        if (json.dashboard_target) {
            when(() => DashboardTargetStore.models.length > 0, () => {
                runInAction(() => {
                    const target_data = json.dashboard_target as { module_id?: string, dashboard_id?: string };
                    const target = DashboardTargetStore.models.find(t => t.module_id === target_data.module_id && t.dashboard_id === target_data.dashboard_id);
                    if (target) {
                        this.dashboard_target_id = target.id as number;
                    }
                })
            });
        }
    }

    @computed
    get typographyTitleStyle() {
        const style = {
            marginLeft: '16px', marginRight: '16px'
        } as any;
        const position = ((this.viewModel.display_options || {}).title || {}).position;
        if (!position || (position === 'top-left')) {
            style.textAlign = 'left';
        } else if (position === 'top-middle') {
            style.textAlign = 'center';
        } else if (position === 'top-right') {
            style.textAlign = 'right';
        } else if (position === 'omit') {
            style.display = 'none';
        }
        return style;
    }

    get defaultCurve() {
        return 'smooth'
    }

    @computed
    get xAxisLabelsFormatter() {
        const target = (this.viewModel.dashboard_target as DashboardTargetModel);
        const formatter_override = target?.ui_hints?.options_override?.xaxis?.labels?.formatter;

        return getXAxisFormatter(formatter_override, this.isNew);
    }

    @computed
    get tooltipFormatter() {
        const target = (this.viewModel.dashboard_target as DashboardTargetModel);
        const formatter_override = target?.ui_hints?.options_override?.tooltip?.formatter;

        return getTooltipFormatter(formatter_override, this.isNew, target?.ui_hints);
    }

    @computed
    get dataLabelsFormatter() {
        const target = (this.viewModel.dashboard_target as DashboardTargetModel);
        const formatter_override = target?.ui_hints?.options_override?.dataLabels?.formatter;

        return getDataLabelFormatter(formatter_override, this.isNew);
    }

    @computed
    get yIsPercentage() {
        return (this.viewModel.dashboard_target as DashboardTargetModel).ui_hints?.ytype === 'percentage'
    }

    @computed
    get xAxisType() {
        switch ((this.viewModel.dashboard_target as DashboardTargetModel).ui_hints?.xtype) {
            case 'datetime':
                return 'datetime'
            case 'number':
                return 'numeric'
            default:
                return undefined
        }
    }

    @computed
    get labels() {
        return (this.viewModel.data.categories || []).map((c: ReturnedCategory) => c.name);
    }

    @computed
    get availableCategories() {
        return ((this.viewModel.dashboard_target as DashboardTargetModel)?.ui_hints?.categories || []) as Array<AvailableCategory>
    }

    @computed
    get availableSeries() {
        return ((this.viewModel.dashboard_target as DashboardTargetModel)?.ui_hints?.series || []) as Array<AvailableSeries>
    }

    @action.bound
    setOverrideColor(id: string, color: string) {
        this.nestedAssignment('display_options.category_colors', [...(this.viewModel.display_options?.category_colors || []).filter((oc: { id: string; }) => {
            return oc.id !== id
        }), {id: id, color: color}])
    }

    chartColor = computedFn(function getDef(this: DashboardComponentModel, id: string, name: string): any {
        const overrideColors = this.viewModel.display_options?.category_colors || []
        const match = overrideColors.find((oc: { id: string; }) => {
            return oc.id === id;
        }) || this.colorTargets.find((oc: AvailableCategory) => {
            return oc.value === id;
        });
        return match?.color || stringToColour(name);
    });

    @computed get chartColors() {
        const target = this.viewModel.dashboard_target as DashboardTargetModel;
        if (target?.ui_hints?.type === 'category_chart') {
            return (this.viewModel.data.categories || []).map((c: ReturnedCategory) => {
                return this.chartColor(c.id, c.name);
            });
        }
        if (target?.ui_hints?.type === 'dual_axis_chart') {
            const data = this.viewModel.data?.series;
            if (isArrayLike(data) && ('id' in data[0])) {
                return data.map((c: ReturnedSeries) => {
                    return this.chartColor(c.id, c.name);
                });
            }
        }
        if (target?.ui_hints?.type === 'type_override') {
            return target?.ui_hints?.options_override?.colors || this.viewModel.data?.colors || null;
        }
        return null;
    }

    @computed
    get colorTargets(): AvailableCategory[] | AvailableSeries[] {
        const target = this.viewModel.dashboard_target as DashboardTargetModel;
        if (target?.ui_hints?.type === 'category_chart') {
            return this.availableCategories;
        }
        if (target?.ui_hints?.type === 'dual_axis_chart') {
            return this.availableSeries;
        }
        return [];
    }

    @computed
    get chartAnnotations() {
        return (this.viewModel.display_options.annotations || []).reduce((annotations: ApexAnnotations, a: Annotation) => {
            const value = Number(a.value);
            const thickness = Number(a.thickness) || 4;
            const color = a.color || '#8B0000';
            const dashSpacing = a.dashed ? (10 * Math.log(thickness)) : 0
            const annotation = {
                strokeDashArray: dashSpacing,
                opacity: 0.6,
                borderColor: color,
                borderWidth: thickness,
                label: {
                    text: a.label,
                    style: {
                        fontSize: `${8 + thickness}px`
                    }
                }
            } as YAxisAnnotations | XAxisAnnotations;
            if (a.type === 'yaxis') {
                (annotation as YAxisAnnotations).y = value;
                annotation.label!.position = 'left';
                annotation.label!.textAnchor = 'start';
                annotations.yaxis!.push(annotation);
            } else {
                (annotation as XAxisAnnotations).x = value;
                annotations.xaxis!.push(annotation);
            }

            return annotations;
        }, {
            xaxis: [] as XAxisAnnotations[],
            yaxis: [] as YAxisAnnotations[],
        });
    }

    @computed
    get graphOptions() {
        const target = this.viewModel.dashboard_target as DashboardTargetModel;

        const options_override = target?.ui_hints?.options_override;

        const options: ApexOptions = {
            dataLabels: {
                enabled: options_override?.dataLabels?.enabled != null ? options_override?.dataLabels?.enabled : this.viewModel.display_options?.dataLabels?.enabled,
                formatter: this.dataLabelsFormatter
            },
            labels: this.labels,
            stroke: {
                curve: this.viewModel.display_options?.line?.stroke?.curve || this.defaultCurve
            },
            xaxis: {
                type: this.xAxisType,
                categories: (this.viewModel.data.categories || []).map((c: ReturnedCategory) => c.name),
                title: {
                    text: options_override?.xaxis?.title?.text || undefined
                },
                labels: {
                    formatter: this.xAxisLabelsFormatter
                }
            },
            yaxis: {
                title: {
                    text: options_override?.yaxis?.title?.text || undefined
                }
            },
            chart: {
                animations: {
                    enabled: false
                },
                stacked: options_override?.chart?.stacked || false,
                stackType: options_override?.chart?.stackType || "normal"
            },
            plotOptions: {
                bar: {
                    horizontal: options_override?.plotOptions?.bar?.horizontal || false
                }
            },
            title: {
                margin: 10,
                style: {
                    fontSize: '24',
                    fontWeight: 400
                }
            },
            tooltip: {
                enabledOnSeries: options_override?.tooltip?.enabledOnSeries || undefined,
                y: {
                    formatter: this.tooltipFormatter
                }
            },
            legend: {
                offsetY: options_override?.legend?.offsetY != null ? options_override?.legend?.offsetY : 30,
                position: options_override?.legend?.position || "bottom",
                horizontalAlign: options_override?.legend?.horizontalAlign || "center",
                customLegendItems: options_override?.legend?.customLegendItems || []
            }
        };

        if (this.chartAnnotations.xaxis.length || this.chartAnnotations.yaxis.length) {
            options.annotations = this.chartAnnotations;
        }

        if (this.yIsPercentage) {
            (options.yaxis as ApexYAxis).min = 0;
            (options.yaxis as ApexYAxis).max = 100;
        }

        if ((this.viewModel.display_options?.title?.position === 'top-left') || (this.viewModel.display_options?.title?.position === undefined)) {
            options.title!.align = 'left';
            options.title!.text = this.viewModel.name;
        } else if (this.viewModel.display_options?.title?.position === 'top-middle') {
            options.title!.align = 'center';
            options.title!.text = this.viewModel.name;
        } else if (this.viewModel.display_options?.title?.position === 'top-right') {
            options.title!.align = 'right';
            options.title!.text = this.viewModel.name;
        }

        if (this.chartColors && this.chartColors.length) {
            options.colors = this.chartColors;
        }

        return options;
    }
}

export const DashboardComponentStore = DashboardComponentModel.store as unknown as DataStore<DashboardComponentModel>;
DashboardComponentStore.setFetchParams({limit: undefined, page: 0, sorted: [], filtered: []});